@@ -1,7 +1,7 @@
 package App::shcompgen;
 
-our $DATE = '2016-10-25'; # DATE
-our $VERSION = '0.24'; # VERSION
+our $DATE = '2016-10-26'; # DATE
+our $VERSION = '0.25'; # VERSION
 
 use 5.010001;
 use strict;
@@ -46,6 +46,7 @@ our %shell_arg = (
     shell => {
         summary => 'Override guessing and select shell manually',
         schema => ['str*', {in=>\@supported_shells}],
+        tags => ['common'],
     },
 );
 our %common_args = (
@@ -71,6 +72,7 @@ user, per-user directory is chosen. Using `--global` or `--per-user` overrides
 that and manually select which.
 
 _
+        tags => ['common'],
     },
 
     bash_global_dir => {
@@ -78,50 +80,60 @@ _
         schema  => ['array*', of => 'str*'],
         default => ['/usr/share/bash-completion/completions',
                     '/etc/bash_completion.d'],
+        tags => ['common'],
     },
     bash_per_user_dir => {
         summary => 'Directory to put completions scripts',
         schema  => ['array*', of => 'str*'],
+        tags => ['common'],
     },
 
     fish_global_dir => {
         summary => 'Directory to put completions scripts',
         schema  => ['array*', of => 'str*'],
         default => ['/usr/share/fish/completions', '/etc/fish/completions'],
+        tags => ['common'],
     },
     fish_per_user_dir => {
         summary => 'Directory to put completions scripts',
         schema  => ['array*', of => 'str*'],
+        tags => ['common'],
     },
 
     tcsh_global_dir => {
         summary => 'Directory to put completions scripts',
-        description => <<'_',
-
-Currently actually hardcoded to `/etc/tcsh/completions`.
-
-_
         schema  => ['array*', of => 'str*'],
         default => ['/etc/tcsh/completions'],
+        tags => ['common'],
     },
     tcsh_per_user_dir => {
         summary => 'Directory to put completions scripts',
-        description => <<'_',
-
-Currently actually hardcoded to `~/.config/tcsh/completions`.
-
-_
         schema  => ['array*', of => 'str*'],
+        tags => ['common'],
     },
 
     zsh_global_dir => {
         summary => 'Directory to put completions scripts',
         schema  => ['array*', of => 'str*'],
         default => ['/usr/local/share/zsh/site-functions'],
+        tags => ['common'],
     },
     zsh_per_user_dir => {
         summary => 'Directory to put completions scripts',
         schema  => ['array*', of => 'str*'],
+        tags => ['common'],
+    },
+
+    helper_global_dir => {
+        summary => 'Directory to put helper scripts',
+        schema  => ['str*'],
+        default => '/etc/shcompgen/helpers',
+        tags => ['common'],
+    },
+    helper_per_user_dir => {
+        summary => 'Directory to put helper scripts',
+        schema  => ['str*'],
+        tags => ['common'],
     },
 );
 
@@ -167,6 +179,8 @@ sub _set_args_defaults {
     $args->{tcsh_per_user_dir} //= ["$ENV{HOME}/.config/tcsh/completions"];
     $args->{zsh_global_dir}    //= ['/usr/local/share/zsh/site-functions'];
     $args->{zsh_per_user_dir}  //= ["$ENV{HOME}/.config/zsh/completions"];
+    $args->{helper_global_dir}   //= '/etc/shcompgen/helpers';
+    $args->{helper_per_user_dir} //= "$ENV{HOME}/.config/shcompgen/helpers";
     [200];
 }
 
@@ -219,6 +233,38 @@ sub _gen_completion_script {
 
     my $header_at_bottom;
     my $script;
+    my @helper_scripts;
+
+    if ($shell ne 'fish' &&
+            ($detres->[3]{'func.completer_type'} // '') eq 'Getopt::Long') {
+        my $progpath = $detres->[3]{'func.program_path'};
+        my $content;
+        require Getopt::Long::Dump;
+        my $dump_res = Getopt::Long::Dump::dump_getopt_long_script(
+            filename => $progpath,
+        );
+        if ($dump_res->[0] != 200) {
+            $log->errorf("Can't dump Getopt::Long script '%s': %s", $progpath, $dump_res);
+            $script = "# Can't dump Getopt::Long script '$progpath': $dump_res->[0] - $dump_res->[2]\n";
+            goto L1;
+        }
+        $content = join(
+            "",
+            "#!$^X\n",
+            "use Getopt::Long::Complete;\n",
+            "my \$spec = ", Data::Dmp::dmp($dump_res->[2]), ";\n",
+            "GetOptions(%\$spec);\n",
+        );
+        $comp = ($args{global} ?
+                     $args{helper_global_dir} : $args{helper_per_user_dir}) .
+            "/$prog";
+        $qcomp = String::ShellQuote::shell_quote($comp);
+        push @helper_scripts, {
+            path => $comp,
+            content => $content,
+        };
+    }
+
     if ($shell eq 'bash') {
         if (defined $args) {
 
@@ -258,7 +304,7 @@ complete -F _|."$prog $qprog".q|
             $script = q|#compdef |.$prog.q|
 _|.$prog.q|() {
     si=$IFS
-    compadd -- $(COMP_LINE=$BUFFER COMP_POINT=$CURSOR |.$qprog.q|)
+    compadd -- $(COMP_LINE=$BUFFER COMP_POINT=$CURSOR |.$qcomp.q|)
     IFS=$si
 }
 _|.$prog.q| "$@"
@@ -310,7 +356,7 @@ _|.$prog.q| "$@"
             $script = "# TODO for type=$type\n";
         }
     } else {
-        die "Sorry, shells other than bash/fish are not supported yet";
+        die "Sorry, shell '$shell' is not supported yet";
     }
 
   L1:
@@ -323,7 +369,13 @@ _|.$prog.q| "$@"
             ($detres->[3]{'func.note'} // ''). "\n$script\n";
     }
 
-    $script;
+    my $i = 0;
+    for (@helper_scripts) {
+        $i++;
+        $script .= "# FRAGMENT id=shcompgen-helper-$i path=$_->{path}\n";
+    }
+
+    ($script, @helper_scripts);
 }
 
 sub _completion_scripts_dirs {
@@ -517,9 +569,10 @@ sub _detect_prog {
             }];
         }
 
-        if ($shell eq 'fish' && $is_perl_script &&
+        if ($is_perl_script &&
                 $content =~ /^\s*((?:use|require)\s+(Getopt::Long))\b/m) {
             return [200, "OK", 1, {
+                "func.program_path" => $progpath,
                 "func.completer_command" => $prog,
                 "func.completer_type" => $2,
                 "func.note"=>"perl use/require statement: $1",
@@ -604,7 +657,7 @@ sub _generate_or_remove {
                 }
             }
 
-            my $script = _gen_completion_script(
+            my ($script, @helper_scripts) = _gen_completion_script(
                 %args, prog => $prog, detect_res => $detres);
             my $comppath = _completion_script_path(
                 %args, prog => $prog, detect_res => $detres);
@@ -627,9 +680,22 @@ sub _generate_or_remove {
             if ($@) {
                 $envres->add_result(500, "Can't write to '$comppath': $@",
                                     {item_id=>$prog0});
-            } else {
-                $envres->add_result(200, "OK", {item_id=>$prog0});
+                next PROG;
             }
+            for my $hs (@helper_scripts) {
+                $log->infof("Writing helper script %s ...", $hs->{path});
+                $written_files{$hs->{path}}++;
+                eval {
+                    write_text($hs->{path}, $hs->{content});
+                    chmod 0755, $hs->{path};
+                };
+                if ($@) {
+                    $envres->add_result(500, "Can't write helper script to '$hs->{path}': $@",
+                                        {item_id=>$prog0});
+                    next PROG;
+                }
+            }
+            $envres->add_result(200, "OK", {item_id=>$prog0});
         } # generate
 
       REMOVE:
@@ -656,13 +722,25 @@ sub _generate_or_remove {
                 next PROG;
             }
             $log->infof("Unlinking %s ...", $comppath);
-            if (unlink $comppath) {
-                $envres->add_result(200, "OK", {item_id=>$prog0});
-                $removed_files{$comppath}++;
-            } else {
+            unless (unlink $comppath) {
                 $envres->add_result(500, "Can't unlink '$comppath': $!",
                                     {item_id=>$prog0});
+                next PROG;
             }
+            # XXX we should only remove helper script if there are no other
+            # shells' completion scripts using this
+            while ($content =~ /^# FRAGMENT id=shcompgen-helper-\d+ path=(.+)/mg) {
+                my $hspath = $1;
+                $log->infof("Unlinking helper script %s ...", $1);
+                unless (unlink $hspath) {
+                    $envres->add_result(500, "Can't unlink helper script '$hspath': $!",
+                                        {item_id=>$prog0});
+                    next PROG;
+                }
+                $removed_files{$hspath}++;
+            }
+            $envres->add_result(200, "OK", {item_id=>$prog0});
+            $removed_files{$comppath}++;
         } # remove
 
     } # for prog0
@@ -759,6 +837,12 @@ sub init {
 
     $dirs = _completion_scripts_dirs(%args);
 
+    if ($global) {
+        push @$dirs, $args{helper_global_dir};
+    } else {
+        push @$dirs, $args{helper_per_user_dir};
+    }
+
     if ($shell eq 'bash') {
         $init_location = $global ? "/etc/bash.bashrc" : "~/.bashrc";
         $init_script = <<_;
@@ -927,6 +1011,8 @@ sub list {
     my $setdef_res = _set_args_defaults(\%args);
     return $setdef_res unless $setdef_res->[0] == 200;
 
+    my $shell = $args{shell};
+
     my @res;
     my $resmeta = {};
     my $dirs = _completion_scripts_dirs(%args);
@@ -937,7 +1023,12 @@ sub list {
             next if $entry eq '.' || $entry eq '..';
 
             # XXX refactor: put to function (_file_to_prog)
-            my $prog = $entry; $prog =~ /\.fish\z/ if $args{shell} eq 'fish';
+            my $prog = $entry;
+            if ($shell eq 'fish') {
+                $prog =~ s/\.fish\z//;
+            } elsif ($shell eq 'zsh') {
+                $prog =~ s/\A_//;
+            }
             next unless $prog =~ $re_progname;
 
             # XXX refactor: put to function (_read_completion_script)
@@ -1024,7 +1115,7 @@ App::shcompgen - Generate shell completion scripts
 
 =head1 VERSION
 
-This document describes version 0.24 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-25.
+This document describes version 0.25 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-26.
 
 =head1 FUNCTIONS
 
@@ -1097,6 +1188,14 @@ By default, if running as root, the global is chosen. And if running as normal
 user, per-user directory is chosen. Using C<--global> or C<--per-user> overrides
 that and manually select which.
 
+=item * B<helper_global_dir> => I<str> (default: "/etc/shcompgen/helpers")
+
+Directory to put helper scripts.
+
+=item * B<helper_per_user_dir> => I<str>
+
+Directory to put helper scripts.
+
 =item * B<prog> => I<array[str]>
 
 Program(s) to generate completion for.
@@ -1130,14 +1229,10 @@ Output completion script to STDOUT.
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C</etc/tcsh/completions>.
-
 =item * B<tcsh_per_user_dir> => I<array[str]>
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C<~/.config/tcsh/completions>.
-
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1221,6 +1316,14 @@ By default, if running as root, the global is chosen. And if running as normal
 user, per-user directory is chosen. Using C<--global> or C<--per-user> overrides
 that and manually select which.
 
+=item * B<helper_global_dir> => I<str> (default: "/etc/shcompgen/helpers")
+
+Directory to put helper scripts.
+
+=item * B<helper_per_user_dir> => I<str>
+
+Directory to put helper scripts.
+
 =item * B<shell> => I<str>
 
 Override guessing and select shell manually.
@@ -1229,14 +1332,10 @@ Override guessing and select shell manually.
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C</etc/tcsh/completions>.
-
 =item * B<tcsh_per_user_dir> => I<array[str]>
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C<~/.config/tcsh/completions>.
-
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1299,6 +1398,14 @@ By default, if running as root, the global is chosen. And if running as normal
 user, per-user directory is chosen. Using C<--global> or C<--per-user> overrides
 that and manually select which.
 
+=item * B<helper_global_dir> => I<str> (default: "/etc/shcompgen/helpers")
+
+Directory to put helper scripts.
+
+=item * B<helper_per_user_dir> => I<str>
+
+Directory to put helper scripts.
+
 =item * B<shell> => I<str>
 
 Override guessing and select shell manually.
@@ -1307,14 +1414,10 @@ Override guessing and select shell manually.
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C</etc/tcsh/completions>.
-
 =item * B<tcsh_per_user_dir> => I<array[str]>
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C<~/.config/tcsh/completions>.
-
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
@@ -1375,6 +1478,14 @@ By default, if running as root, the global is chosen. And if running as normal
 user, per-user directory is chosen. Using C<--global> or C<--per-user> overrides
 that and manually select which.
 
+=item * B<helper_global_dir> => I<str> (default: "/etc/shcompgen/helpers")
+
+Directory to put helper scripts.
+
+=item * B<helper_per_user_dir> => I<str>
+
+Directory to put helper scripts.
+
 =item * B<prog> => I<array[str]>
 
 Program(s) to remove completion script of.
@@ -1390,14 +1501,10 @@ Override guessing and select shell manually.
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C</etc/tcsh/completions>.
-
 =item * B<tcsh_per_user_dir> => I<array[str]>
 
 Directory to put completions scripts.
 
-Currently actually hardcoded to C<~/.config/tcsh/completions>.
-
 =item * B<zsh_global_dir> => I<array[str]> (default: ["/usr/local/share/zsh/site-functions"])
 
 Directory to put completions scripts.
